/* Program for Marklin Turntable (TT) 7186
 *  Function addresses: 
 *                      225 	0 - Red (curved) Set counter to one (1).
 *                            1 - Green (straight) Increase counter by one
 *                      226 	2 - Red (curved) Clockwise rotation (180°)
 *          	                3 - Green (straight) Anti-Clockwise rotation (180°)
 *                      227 	4 - Red (curved) Move one step to the Clockwise.
 *          	                5 - Green (straight) Move one step to the Anti-Clockwise.
 *                      228 	6 - Red (curved) Set rotation direction to Clockwise.
 *                          	7 - Green (straight) Set rotation direction to (Anti-clockwise)
 *
 *  Function addresses: 229 - 232 >  places 0 to 7 of the turntable
 *
 *  For Rocrail an extra action: after code 228 (Green or Red) is received, 
 *                          code 226 (Red) turns the TT 180°/ Clock- or Anti-clockwise
 *
 *  Pin connections of Arduino Pro mini
 *  Pin2: (D2) D0 of MOT-1   3-bits data of Motorola IC1
 *  Pin3: (D3) D1 of MOT-1       ""
 *  Pin4: (D4) D2 of MOT-1       ""
 *  Pin5: (D5) D0 of MOT-2   3-bits data of Motorola IC2
 *  Pin6: (D6) D1 of MOT-2       ""
 *  Pin7: (D7) D2 of MOT-2       ""
 *  Pin8: (D8) D3 of MOT-1   if 1: Motorola IC1 is active
 *  Pin9: (D9) D3 of MOT-2   if 1: Motorola IC2 is active
 *  Pin10: (D10) 1A of DC4511 (BCD to 7-segment)
 *  Pin11: (D11) 1B of DC4511        ""   
 *  Pin12: (D12) 1C of DC4511        ""
 *  Pin13: (D13) REED Relay (S88)
 *  Pin14: (D14/A0): Optocoupler Digital Signal   Low Active
 *  Pin15: (D15/A1): Optocoupler (Motor)          Low Active
 *  Pin16: (D16/A2): Relay Pulse
 *  Pin17: (D17/A3): Relay Direction
 *  Pin18: (D18/A4): ExtraInfoPin. If Pin 18 is LOW, extra Info is shown on the 7-Segment Display
 *  Pin19: (D19/A5): 1D of DC4511    (BCD to 7-segment) 
 *  Pin20: (A6): NC
 *  Pin21: (A7): NC
 *  
 *  eeprom:
       addr + 0: posArr
       EEPROM.write(addr+0,posArr);
       posArr = EEPROM.read(addr+0);
       
 by Chris Burger
 written January 2023  v.15
 */

// include the library code:
#include <EEPROM.h>

const byte inputMot01 =  2;  // 3 pins of Mot-IC 1  - PD2
const byte inputMot02 =  3;  //                     - PD3
const byte inputMot03 =  4;  //                     - PD4
const byte inputMot11 =  5;  // 3 pins of Mot-IC 2  - PD5
const byte inputMot12 =  6;  //                     - PD6
const byte inputMot13 =  7;  //                     - PD7
const byte inputMot04 =  8;   // Mot-IC 1 active
const byte inputMot14 =  9;   // MOT-IC 2 active
const byte DC4511_1   = 10;   //                    - PB2
const byte DC4511_2   = 11;   //                    - PB3
const byte DC4511_3   = 12;   //                    - PB4
const byte RelayS88   = 13;
const byte CurrentM   = 14; // low active
const byte VoltageDS  = 15; // low active
const byte RelayP     = 16;
const byte RelayD     = 17;
const byte ExtraInfoPin  = 18;
const byte DC4511_4   = 19; 

byte arrayLR[10]     = {0, 1, 2, 3, 4, 5, 6, 7, 1, 6};
byte stepsClock[10]  = {1, 1, 1, 1, 1, 1, 1, 2, 5, 2}; 
byte stepsAClock[10] = {2, 1, 1, 1, 1, 1, 1, 1, 2, 5}; 

//boolean Current;
byte i;
byte p;
byte pos;   // postion of turntable
byte posArr;   // position in arrayLR
byte steps;
boolean saveBytes;  // if 0 posArr has to be stored in EEPROM
int pulseTime = 700;  //  PulseTime relay in Turntable
int checkTime = 200;  //  Time to check interruption
boolean currentMot;
boolean digSign;
boolean wait1;
boolean mot1_Active;   
boolean mot2_Active;
byte port_D;
byte actionMot1;           // 0 - 7: action of Motorola IC1
byte actionMot1_old;
byte actionMot2;           // 0 - 7: action of Motorola IC2
unsigned long startMillis;
unsigned long endMillis;
int milSeconds;
boolean stopWhile;
const int  addr = 0 ;
boolean extraInfo;
byte Blank = 15;  // makes display BLANK

void setup(){
    pinMode (inputMot01, INPUT);
    pinMode (inputMot02, INPUT);
    pinMode (inputMot03, INPUT);
    pinMode (inputMot04, INPUT);
    pinMode (inputMot11, INPUT);
    pinMode (inputMot12, INPUT);
    pinMode (inputMot13, INPUT);
    pinMode (inputMot14, INPUT);   
    pinMode (DC4511_1, OUTPUT);
    pinMode (DC4511_2, OUTPUT);  
    pinMode (DC4511_3, OUTPUT);
    pinMode (DC4511_4, OUTPUT);
    pinMode (RelayS88, OUTPUT);
    pinMode (RelayP, OUTPUT);
    pinMode (RelayD, OUTPUT);
    pinMode (CurrentM, INPUT_PULLUP);
    pinMode (VoltageDS, INPUT_PULLUP);
    pinMode (ExtraInfoPin, INPUT_PULLUP);   // reads jumper E
          }
          
void loop(){
            // switch off Relays
    digitalWrite (RelayD,LOW);  // RelayD off
    digitalWrite (RelayS88,LOW);  // RelayS88 off
    digitalWrite (RelayP,LOW);  // RelayP off
    extraInfo = digitalRead(ExtraInfoPin);
    
        if (extraInfo == LOW)
          {
     // show the digits 0 - 9 on 7-Segment display for check
              for (i=0; i < 10; i++)
             {
               SevenSegment (i);
                delay(300);
               SevenSegment (Blank);   // display BLANK for 150 msec
                delay(150);
            }
          }
             saveBytes  = 1; 
    // show postion the turntable op 7-Segment display
    // information is stored in EEPROM
                  posArr = EEPROM.read(addr+0);
                        if (posArr > 9)          
                          { posArr = 6;
                             saveBytes = 0; }
                   pos = arrayLR[posArr] ; 
                    SevenSegment (pos + 1);
                            
                 while(1)
                {      
                  digSign = digitalRead(VoltageDS);   // Low Active
              // Check if DigitalSignal is present and posArr is not saved in EEprom.
                    if (digSign == 1   &&  saveBytes == 0)
                     {
                         EEPROM.write(addr+0,posArr);
                           saveBytes = 1;
                     }
            extraInfo = digitalRead(ExtraInfoPin);        
       // Check if DigitalSignal is present.
              digSign = digitalRead(VoltageDS);
         while  (digSign == 0)  // digital Voltage present  
          {
            // Check if a Signal is sent to the decoder
           mot1_Active = digitalRead(inputMot04);
            if (mot1_Active == 1)  // has Motorola IC1 received a code?
              {  
                  delay(5);
               // Check which action is active!!            
                   actionMot1= PIND ;
                    actionMot1 =  actionMot1 >> 2 ;
                     actionMot1 =  actionMot1 & 0x07;         

               if (actionMot1 == 6 || actionMot1 == 7) // @@@@@
                  {                               
                        pos = arrayLR[posArr]; 
               // Check three actions:
               //     After code 228r or 228g the decoder is receiving code 226r or 226g (180 degree turning)
               //     After code 228r of 228g the decoder is receiving code 229 to 232 (moving to another track)
               //     After code 228r of 228g and after a pause of 6 sec the directionrelay is switched on or off.

                             if (actionMot1 == 6) 
                             { digitalWrite (RelayD,HIGH); } // RelayD on/Clockwise turning
                             else
                             { digitalWrite (RelayD,LOW); }  // RelayD off
                             
                                 // actionMot1 has to be usd later on
                                      actionMot1_old = actionMot1;
                                       mot1_Active = digitalRead(inputMot04);
                                        mot2_Active = digitalRead(inputMot14); 
                                   // wait until Motorola IC1 is in-active or IC2 is active                               
                                      while (mot1_Active == 1 && mot2_Active == 0)
                                      {
                                    mot1_Active = digitalRead(inputMot04);
                                    mot2_Active = digitalRead(inputMot14); 
                                      }
                                     if (mot2_Active == 1)
                                     {
                                       (mot1_Active = 0);
                                     }
                                      else
                                     { 
                                      
                   // check if MotIC1 or MotIC2 is active. Do this for 6 seconds
                            startMillis = millis(); 
                              endMillis = millis();
                            mot1_Active = digitalRead(inputMot04);  
                            mot2_Active = digitalRead(inputMot14);  
                            
                          while (((mot1_Active == 0) && (mot2_Active == 0)) && (endMillis - startMillis) < 6000)
                              {        
                                    endMillis = millis();
                                  mot1_Active = digitalRead(inputMot04);   
                                  mot2_Active = digitalRead(inputMot14);                
                              }
                                    
                              if (mot1_Active == 1)
                              {
                             // Check which action is active!! 
                            actionMot1 =  PIND ;
                            actionMot1 =  actionMot1 >> 2 ;
                            actionMot1 =  actionMot1 & 0x07;
                              }
                               }
                           if   (mot1_Active == 1 && (actionMot1 == 2 || actionMot1 == 3) ) //@@@@@
                         // After code 228r or 228g the decoder is receiving code 226r or 226g (180 degree turning)
                            {
                               digitalWrite (RelayS88,HIGH);  // RelayS88 on  
                                // if extraInfo is HIGH: no extra info on display 
                                    if (extraInfo == LOW)
                                    {
                                      // Warning Signal
                                    for (i=0; i < 4; i++)
                                  {
                                    SevenSegment (8);
                                          delay(150);
                                       // display BLANK for 100 msec
                                    SevenSegment (Blank);  
                                          delay(100);
                                  }
                                  
                                // display 6 = Clockwise or 7 = Anti-Clockwise
                                     SevenSegment (actionMot1_old);  
                                     delay (2000); 
                                     SevenSegment (Blank);   // display BLANK
                                        delay(500);
                                        // Warning Signal
                                    for (i=0; i < 4; i++)
                                  {
                                    SevenSegment (8);
                                          delay(150);
                                       // display BLANK for 100 msec
                                    SevenSegment (Blank);  
                                          delay(100);
                                  }
                                // display 2 = Clockwise or 3 = Anti-Clockwise
                                     SevenSegment (actionMot1);  
                                     delay (2000); 
                                     SevenSegment (Blank);   // display BLANK
                                        delay(500);

                                    }
                                            // turn 180 degrees
                                            for (i = 0; i < 8; i++)
                                      {
                                          // switch the motor on     
                                          digitalWrite (RelayP,HIGH);
                                                delay(pulseTime);
                                          digitalWrite (RelayP,LOW);
                                            SevenSegment (8 - i); 
  
                                        // check the current to the motor  
                                       currentMot = digitalRead(CurrentM); 
                                       
                                     // wait until the current to the turntable in zero
                                          while (currentMot == LOW)   
                                           {
                                             startMillis = millis(); 
                                             endMillis = millis();                                       
                                     // check if Motor is active 
                                        currentMot = digitalRead(CurrentM);
                                             // check if Motor is really 100 - 300 msec off
                                         while (currentMot == HIGH && (endMillis - startMillis) < checkTime)
                                             {   endMillis = millis();
                                                 currentMot = digitalRead(CurrentM); 
                                             }
                                         }
                                      }
                                      //   display 0
                                       SevenSegment (0);  
                                          delay (500); 
                                         SevenSegment (Blank);   // display BLANK
                                        delay(200);
                                      //   display position turntable
                                             pos = arrayLR[posArr] ; 
                                              SevenSegment (pos + 1);
                                   digitalWrite (RelayD,LOW); // RelayD off
                                   digitalWrite (RelayS88,LOW);  // RelayS88 off
                                   actionMot1 = 7;  // to prevent to start action 2 or 3 again
                                   mot2_Active = 0; // to prevent to start action 6 or 7
                            }
                                       
                           if   (mot2_Active == 1)                   //@@@@@
                   //     After code 228r of 228g the decoder is receiving code 229 to 232 (moving to another track)
                               {  
                               digitalWrite (RelayS88,HIGH);  // RelayS88 on
                                 actionMot1 = actionMot1_old;
                                 // Check which to which position the turntable has to go  
                                 actionMot2 = PIND;
                                   actionMot2 =  actionMot2 >> 5 ;
                                    actionMot2 = actionMot2 & 0x07 ; 
                                      pos = actionMot2;

                                    if (extraInfo == LOW)
                                    {
                                      // Warning Signal
                                    for (i=0; i < 4; i++)
                                  {
                                    SevenSegment (8);
                                          delay(150);
                                       // display BLANK for 100 msec
                                    SevenSegment (Blank);  
                                          delay(100);
                                  }
                                // display 6 = Clockwise or 7 = Anti-Clockwise
                                     SevenSegment (actionMot1);  
                                     delay (2000); 
                                     SevenSegment (Blank);   // display BLANK
                                        delay(500);
                                // display number of desired place of turntable                         
                                    SevenSegment (actionMot2 + 1);  
                                     delay (2000); 
                                     SevenSegment (Blank);    // display BLANK
                                        delay(500);
                                    }
                                      if (actionMot2 == posArr)  // to check 
                                      {  
                                        steps = 0; // actionMot2 is equal to posArr
                                      }             // to turn 0 degrees
                                        else               
                                        {
                                      steps = 0;
                                      stopWhile = 0;
                                    // direction of turninig (6 = clockwise)
                                 if (actionMot1 == 6)  // clockwise
                                 {
                                // calculate how many steps the TT had to go
                                       while (stopWhile == 0)
                                     {
                                    if (actionMot2 == posArr)
                                         {
                                          stopWhile = 1;
                                         }
                                         else
                                         {
                                             steps = steps + stepsClock[posArr];
                                              posArr++;
                                            if (posArr > 9)                                 
                                           { posArr = 0; }
                                          }
                                        }
                                      }
                                      else
                                      {           // Anti-clockwise
                                        while (stopWhile == 0)
                                     {
                                        if (actionMot2 == posArr)
                                         {
                                          stopWhile = 1;
                                         }
                                         else
                                         {
                                             steps = steps + stepsAClock[posArr];
                                              posArr--;
                                        if (posArr == 255)    // 255 is equal to -1
                                                    { posArr = 9; }
                                          }
                                        }   
                                      }
                                    }
                                           if (steps > 8)
                                           {   steps = steps - 8; }
                                          for (i = steps; i > 0; i--)
                                      {         SevenSegment (i);
                                          // switch the motor on         
                                          digitalWrite (RelayP,HIGH);
                                                  delay(pulseTime);
                                          digitalWrite (RelayP,LOW);
                                          
                                        // check the current to the motor  
                                       currentMot = digitalRead(CurrentM); 
                                       
                                     // wait until the current to the turntable in zero
                                          while (currentMot == LOW)   
                                         {
                                    startMillis = millis(); 
                                    // check if Motor is active 
                                       endMillis = millis();
                                        currentMot = digitalRead(CurrentM);
                                           // check if Motor is really 100 - 300 msec off
                                         while (currentMot == HIGH && (endMillis - startMillis) < checkTime)
                                             {   endMillis = millis();
                                                 currentMot = digitalRead(CurrentM); 
                                             }
                                         }
                                      }    
                                      //   display 0
                                       SevenSegment (0);  
                                          delay (500); 
                                         SevenSegment (Blank);   // display BLANK
                                        delay(200);
                                    // display position of turntable 
                                       pos = arrayLR[posArr] ;       
                                       SevenSegment (pos + 1);
                                     digitalWrite (RelayS88,LOW);  // RelayS88 off
                                            saveBytes  = 0;
                                      digitalWrite (RelayD,LOW);  // RelayD off
                                         }
                                         
                                      //   No second code was received.
                                        if (mot2_Active == 0 && mot1_Active == 0 ) //@@@@@
                      //     After code 228r of 228g and after a pause of 6 sec the directionrelay is switched on or off.
                                        {
                                            if (extraInfo == LOW)
                                    {
                                       actionMot1 = actionMot1_old;
                                      // Warning Signal
                                    for (i=0; i < 4; i++)
                                  {
                                    SevenSegment (8);
                                          delay(150);
                                       // display BLANK for 100 msec
                                    SevenSegment (Blank);  
                                          delay(100);
                                  }
                                // display 6 = Clockwise or 7 = Anti-Clockwise
                                     SevenSegment (actionMot1);  
                                     delay (2000); 
                                     SevenSegment (Blank);   // display BLANK
                                        delay(500);
                                    }
                                          pos = arrayLR[posArr] ; 
                                          SevenSegment (pos + 1);
                                        }
                                          actionMot1 = 7;  // to prevent to start action 2 or 3 again
                                       }
                                     
                                       // One step to the left or to the right         @@@@@
                                    if (actionMot1 == 4 || actionMot1 == 5) 
                                       {
                                        digitalWrite (RelayS88,HIGH);  // RelayS88 on  
                                         if (actionMot1 == 5) 
                                         {
                                          digitalWrite (RelayD,LOW);  // RelayD off
                                                    posArr--;
                                                   if (posArr == 255)    // 255 is equal to -1
                                                   { posArr = 9; }
                                                   pos = arrayLR[posArr];
                                         }
                                              else
                                          {
                                            digitalWrite (RelayD,HIGH);  // RelayD on
                                                      posArr++;
                                                    if (posArr > 9)
                                                   { posArr = 0; }
                                                   pos = arrayLR[posArr];
                                          }
                                          
                                         if (extraInfo == LOW)
                                         {
                                         // Warning Signal
                                         for (i=0; i < 4; i++)
                                         {
                                         SevenSegment (8);
                                          delay(150);
                                       // display BLANK for 100 msec
                                           SevenSegment (Blank);  
                                          delay(100);
                                         }
                                // display 4 = One step Clockwise or 5 = one step Anti-Clockwise
                                     SevenSegment (actionMot1);  
                                     delay (2000); 
                                     SevenSegment (Blank);   // display BLANK
                                        delay(500);
                                        }
                           
                                          // switch the motor on   
                                           digitalWrite (RelayP,HIGH);
                                                  delay(pulseTime);
                                          digitalWrite (RelayP,LOW);
                                            SevenSegment (1); 
                                        // check the current to the motor  
                                       currentMot = digitalRead(CurrentM); 
                                       
                                     // wait until the current to the turntable in zero
                                          while (currentMot == LOW)   
                                           {
                                             startMillis = millis(); 
                                    // check if Motor is active 
                                       endMillis = millis();
                                        currentMot = digitalRead(CurrentM);
                                             // check if Motor is really 100 - 300 msec off
                                         while (currentMot == HIGH && (endMillis - startMillis) < checkTime)
                                             {   endMillis = millis();
                                                 currentMot = digitalRead(CurrentM); 
                                             }
                                         }
                                          SevenSegment (0);  
                                           delay (500); 
                                         SevenSegment (Blank);   // display BLANK
                                          delay(200);
                                    // display position of turntable 
                                       pos = arrayLR[posArr] ;       
                                       SevenSegment (pos + 1); 
                                        digitalWrite (RelayS88,LOW);  // RelayS88 off
                                              saveBytes  = 0;
                                         digitalWrite (RelayD,LOW);  // RelayD off
                                              SevenSegment (pos + 1);
                                       }   
                                  
                                      // 180 degrees turn around (8 steps)  3 = AntiClock and 2 is Clockwise @@@@@
                                            if (actionMot1 == 2 || actionMot1 == 3) 
                                         {
                                           digitalWrite (RelayS88,HIGH);  // RelayS88 on
                                          if (extraInfo == LOW)
                                          {
                                              // Warning Signal
                                               for (i=0; i < 4; i++)
                                        {
                                            SevenSegment (8);
                                              delay(250);
                                            SevenSegment (Blank);    // display BLANK for 150 msec
                                              delay(150);
                                       }
                                    // display 2 or 3: 180 degrees turnung
                                       SevenSegment (actionMot1);  
                                        delay (2000); 
                                       SevenSegment (Blank); // display BLANK
                                        delay(500);
                                          }
                                        SevenSegment (8); 
                                           if (actionMot1 == 3) 
                                         {  
                                          digitalWrite (RelayD,HIGH);  // RelayD on
                                          }
                                              else
                                          {
                                            digitalWrite (RelayD,LOW);  // RelayD off
                                          }
                                   
                                 
                                            for (i = 0; i < 8; i++)   // 180 degrees is 8 steps
                                      {
                                          // switch the motor on     
                                          digitalWrite (RelayP,HIGH);
                                                delay(pulseTime);
                                          digitalWrite (RelayP,LOW);
                                            SevenSegment (8 - i); 
  
                                        // check the current to the motor  
                                       currentMot = digitalRead(CurrentM); 
                                       
                                     // wait until the current to the turntable in zero
                                          while (currentMot == LOW)   
                                           {
                                             startMillis = millis(); 
                                    // check if Motor is active 
                                       endMillis = millis();
                                        currentMot = digitalRead(CurrentM);
                                             // check if Motor is really 100 - 300 msec off
                                         while (currentMot == HIGH && (endMillis - startMillis) < checkTime)
                                             {   endMillis = millis();
                                                 currentMot = digitalRead(CurrentM); 
                                             }
                                         }
                                      }
                                     //   display 0
                                       SevenSegment (0);  
                                          delay (500); 
                                         SevenSegment (Blank);   // display BLANK
                                         delay(200);
                                      //   display position turntable
                                             pos = arrayLR[posArr] ; 
                                              SevenSegment (pos + 1);
                                       digitalWrite (RelayD,LOW);  // RelayD off
                                       digitalWrite (RelayS88,LOW);  // RelayS88 off                                 
                                        }
                                          // Counter reset and Counter plus 1         @@@@@
                                              if (actionMot1 == 0 || actionMot1 == 1) 
                                         {
                                                   if (actionMot1 == 0) 
                                         {  
                                            posArr = 0;  // reset counter
                                         }
                                         else
                                          {
                                            posArr++;  // increase counter posArr by 1
                                               if (posArr > 9)                                      
                                                  { posArr = 0; }
                                          } 
                                                 pos = arrayLR[posArr];                                 
                                        // wait till pin inputMot04 of Motorola IC1 is not active
                                            mot1_Active = digitalRead(inputMot04);
                                               while ( mot1_Active == 1)
                                              { 
                                                  mot1_Active = digitalRead(inputMot04);
                                              }
                                                // delay(50);  
                                                 SevenSegment (pos + 1);   
                                                 saveBytes  = 0;                                  
                                            }
                                        }
                                          // Check if DigitalSignal is present.
                                           digSign = digitalRead(VoltageDS);                                
                                     }
                                 }
                              }
     
  
             // declaration function SevenSegment
         void SevenSegment (byte pos1) 
         {
          if (pos1 > 7)
             digitalWrite (19,HIGH);  // 19 = DC4511_4
               else
             digitalWrite (19,LOW);
          byte Port_B;   // pin DC4511_1 (10); DC4511_2 (11); DC4511_3 (12) 
          pos1 = pos1 << 2;
          pos1 = pos1 & 0x1C;
          Port_B = PINB;
          Port_B = Port_B & 0xE3;
          Port_B = Port_B | pos1;
          PORTB = Port_B;

         }
